"""
An easy-to-use and lightweight API wrapper for the Censys Search Engine (censys.io).
"""

__author__ = "Censys Team"
__email__ = "support@censys.io"
__copyright__ = "Copyright 2020 Censys, Inc."
__version__ = "1.1.1"
__license__ = "Apache License, Version 2.0"

__all__ = ["asm", "certificates", "data", "exceptions", "ipv4", "websites"]



"""
Interact with the Censys Search IPv4 API.
"""

from censys.api import CensysSearchAPI


class CensysIPv4(CensysSearchAPI):
    """
    Interacts with the IPv4 index.
    """

    INDEX_NAME = "ipv4"
    """Name of Censys Index."""



"""
Interact with the config file.
"""

import os
from pathlib import Path
from configparser import ConfigParser, NoOptionError

from censys import __version__

DEFAULT = "DEFAULT"

xdg_config_path = os.path.join(str(Path.home()), ".config")
censys_path = os.path.join(xdg_config_path, "censys")
config_path = os.path.join(censys_path, "censys.cfg")

default_config = {
    "version": __version__,
    "api_id": "",
    "api_secret": "",
    "asm_api_key": "",
}


def write_config(config):
    """
    Writes config to file.

    Args:
        config: Configuration to write.
    """

    with open(config_path, "w") as configfile:
        config.write(configfile)


def get_config():
    """
    Reads and returns config.
    """

    config = ConfigParser()
    if not os.path.isdir(xdg_config_path):
        os.mkdir(xdg_config_path)
    if not os.path.isdir(censys_path):
        os.mkdir(censys_path)
    if not os.path.exists(config_path):
        config[DEFAULT] = default_config
        with open(config_path, "w") as configfile:
            config.write(configfile)
    config.read(config_path)
    check_config(config)
    if config.get(DEFAULT, "version") != __version__:
        config.set(DEFAULT, "version", __version__)
        write_config(config)
    return config


def check_config(config):
    """
    Checks config against default config for fields.

    Args:
        config: Configuration to write.
    """

    for key in default_config:
        try:
            config.get(DEFAULT, key)
        except NoOptionError:
            config.set(DEFAULT, key, default_config.get(key))
    write_config(config)



"""
Exceptions for Censys.
"""

from typing import Optional, Dict, Type


class CensysException(Exception):
    """
    Base Exception for Censys.
    """


class CensysCLIException(CensysException):
    """
    Exception raised when the CLI is passed invalid arguments.
    """


class CensysMissingApiKeyException(CensysException):
    """
    Exception raised when there is no provided ASM API key.
    """


class CensysAPIException(CensysException):
    """
    Base Exception for Censys API's.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        status_code: int,
        message: str,
        body: Optional[str] = None,
        const: Optional[str] = None,
        error_code: Optional[int] = None,
        details: Optional[str] = None,
    ):
        self.status_code = status_code
        self.message = message
        self.body = body
        self.const = const
        self.error_code = error_code
        self.details = details
        super().__init__(self.message)


class CensysSearchException(CensysAPIException):
    """
    Base Exception for the Censys search API.
    """

    def __repr__(self):
        return "%i (%s): %s" % (self.status_code, self.const, self.message or self.body)

    __str__ = __repr__


class CensysAsmException(CensysAPIException):
    """
    Base Exception for the Censys ASM API.
    """

    def __repr__(self):
        return "%i (Error Code: %i), %s. %s" % (
            self.status_code,
            self.error_code,
            self.message,
            self.details,
        )

    __str__ = __repr__


class CensysRateLimitExceededException(CensysSearchException):
    """
    Exception raised when your Censys rate limit has been exceeded.
    """


class CensysNotFoundException(CensysSearchException):
    """
    Exception raised when the resource requested is not found.
    """


class CensysUnauthorizedException(CensysSearchException):
    """
    Exception raised when your Censys account doesn't have
    access to the requested resource.
    """


class CensysJSONDecodeException(CensysSearchException):
    """
    Exception raised when the resource requested is not valid JSON.
    """


class CensysInvalidRequestException(CensysAsmException):
    """
    Exception raised when the HTTP request is invalid.
    """


class CensysInvalidAuthTokenException(CensysAsmException):
    """
    Exception raised when the auth token is invalid.
    """


class CensysInvalidAPIKeyException(CensysAsmException):
    """
    Exception raised when the ASM API key is invalid.
    """


class CensysTooManyRequestsException(CensysAsmException):
    """
    Exception raised when the allowed requests bandwidth is exceeded.
    """


class CensysAppDownForMaintenanceException(CensysAsmException):
    """
    Exception raised when the ASM API is down for maintenance.
    """


class CensysInvalidSeedDataException(CensysAsmException):
    """
    Exception raised when the seed data is invalid.
    """


class CensysAssociatedAssetsThresholdWarningException(CensysAsmException):
    """
    Exception raised when the associated asset count is within the warning threshold.
    """


class CensysTooManyInputNodesException(CensysAsmException):
    """
    Exception raised when there are too many input nodes.
    """


class CensysSeedNotFoundException(CensysAsmException):
    """
    Exception raised when the requested seed can not be found.
    """


class CensysNotASeedException(CensysAsmException):
    """
    Exception raised when the requested resource is not a seed.
    """


class CensysNeedConfirmationToRemoveParentSeedsException(CensysAsmException):
    """
    Exception raised when confirmation is needed to remove seeds with children.
    """


class CensysCannotRemoveNonExistentSeedsException(CensysAsmException):
    """
    Exception raised when trying to remove non existent seed nodes.
    """


class CensysCannotRemoveNonSeedsException(CensysAsmException):
    """
    Exception raised when trying to remove non seed nodes.
    """


class CensysInvalidSeedTypeException(CensysAsmException):
    """
    Exception raised when the seed type is invalid.
    """


class CensysInvalidLogbookCursorException(CensysAsmException):
    """
    Exception raised when the logbook cursor is invalid.
    """


class CensysPageNumberOutOfRangeException(CensysAsmException):
    """
    Exception raised when the requested page number is out of range [1 - totalPages].
    """


class CensysInvalidPageSizeException(CensysAsmException):
    """
    Exception raised when the page size is invalid.
    """


class CensysHostNotFoundException(CensysAsmException):
    """
    Exception raised when the requested host is not found.
    """


class CensysInvalidIPv4AddressException(CensysAsmException):
    """
    Exception raised when the IPv4 address is invalid.
    """


class CensysInvalidCommentException(CensysAsmException):
    """
    Exception raised when the comment is invalid.
    """


class CensysCommentNotFoundException(CensysAsmException):
    """
    Exception raised when the requested comment is not found.
    """


class CensysInvalidColorException(CensysAsmException):
    """
    Exception raised when the specified color is invalid.
    """


class CensysTagHasTrailingOrLeadingWhitespaceException(CensysAsmException):
    """
    Exception raised when the specified tag has trailing or leading whitespace.
    """


class CensysTagIsEmptyStringException(CensysAsmException):
    """
    Exception raised when the specified tag is an empty string.
    """


class CensysTagLabelsDifferOnlyInCasingException(CensysAsmException):
    """
    Exception raised when the specified tag differs from an existing tag in only casing.
    """


class CensysTagLabelTooLongException(CensysAsmException):
    """
    Exception raised when the specified tag label is too long.
    """


class CensysTagColorTooLongException(CensysAsmException):
    """
    Exception raised when the specified tag color is too long.
    """


class CensysCannotCreateTagWithNewColorException(CensysAsmException):
    """
    Exception raised when the specified tag cannot be created with a new color.
    """


class CensysTagColorHasTrailingOrLeadingWhitespaceException(CensysAsmException):
    """
    Exception raised when the specified tag color has trailing or leading whitespace.
    """


class CensysCertificateNotFoundException(CensysAsmException):
    """
    Exception raised when the certificate is not found.
    """


class CensysDomainNotFoundException(CensysAsmException):
    """
    Exception raised when the domain is not found.
    """


class CensysExceptionMapper:
    """
    Map status code to Exception for the ASM and Search API.
    """

    ASM_EXCEPTIONS: Dict[int, Type[CensysAsmException]] = {
        10008: CensysInvalidRequestException,
        10002: CensysInvalidAuthTokenException,
        10001: CensysInvalidAPIKeyException,
        10039: CensysTooManyRequestsException,
        10029: CensysAppDownForMaintenanceException,
        10007: CensysInvalidSeedDataException,
        10017: CensysAssociatedAssetsThresholdWarningException,
        10016: CensysTooManyInputNodesException,
        10014: CensysSeedNotFoundException,
        10015: CensysNotASeedException,
        10013: CensysNeedConfirmationToRemoveParentSeedsException,
        10012: CensysCannotRemoveNonExistentSeedsException,
        10011: CensysCannotRemoveNonSeedsException,
        10038: CensysInvalidSeedTypeException,
        10040: CensysInvalidLogbookCursorException,
        10051: CensysPageNumberOutOfRangeException,
        10050: CensysInvalidPageSizeException,
        10018: CensysHostNotFoundException,
        10021: CensysInvalidIPv4AddressException,
        10054: CensysInvalidCommentException,
        10055: CensysCommentNotFoundException,
        10037: CensysInvalidColorException,
        10025: CensysTagHasTrailingOrLeadingWhitespaceException,
        10026: CensysTagIsEmptyStringException,
        10027: CensysTagLabelsDifferOnlyInCasingException,
        10028: CensysTagLabelTooLongException,
        10034: CensysTagColorTooLongException,
        10035: CensysCannotCreateTagWithNewColorException,
        10036: CensysTagColorHasTrailingOrLeadingWhitespaceException,
        10020: CensysCertificateNotFoundException,
        10019: CensysDomainNotFoundException,
    }
    """Map of status code to ASM Exception."""

    SEARCH_EXCEPTIONS: Dict[int, Type[CensysSearchException]] = {
        401: CensysUnauthorizedException,
        403: CensysUnauthorizedException,
        404: CensysNotFoundException,
        429: CensysRateLimitExceededException,
    }
    """Map of status code to Search Exception."""



"""
Interact with the Censys Search Certificate API.
"""

from typing import List

from censys.api import CensysSearchAPI


class CensysCertificates(CensysSearchAPI):
    """
    Interacts with the Certificates index.
    """

    INDEX_NAME = "certificates"
    """Name of Censys Index."""
    MAX_PER_BULK_REQUEST = 50
    """Max number of bulk requests."""

    def __init__(self, *args, **kwargs):
        CensysSearchAPI.__init__(self, *args, **kwargs)
        self.bulk_path = f"/bulk/{self.INDEX_NAME}"

    def bulk(self, fingerprints: List[str]) -> dict:
        """
        Requests data in bulk.

        Args:
            fingerprints (List[str]): List of certificate SHA256 fingerprints.

        Returns:
            dict: Search results from an API query.
        """

        result = dict()
        start = 0
        end = self.MAX_PER_BULK_REQUEST
        while start < len(fingerprints):
            data = {"fingerprints": fingerprints[start:end]}
            result.update(self._post(self.bulk_path, data=data))
            start = end
            end += self.MAX_PER_BULK_REQUEST

        return result



"""
Interact with the Censys Search Data API.
"""

from censys.api import CensysSearchAPI


class CensysData(CensysSearchAPI):
    """
    Interacts with the Data index.
    For more details, see our documentation: https://censys.io/api/v1/docs/data
    """

    _PREFIX = "/data"
    """Endpoint prefix."""

    def get_series(self) -> dict:
        """
        Get data on the types of scans we regularly perform (series).

        Returns:
            dict: The result set returned.
        """

        return self._get(self._PREFIX)

    def view_series(self, series_id: str) -> dict:
        """
        Get data on a specific series.

        Args:
            series_id (str): The ID of the series.

        Returns:
            dict: The result set returned.
        """

        path = "/".join([self._PREFIX, series_id])
        return self._get(path)

    def view_result(self, series_id: str, result_id: str) -> dict:
        """
        View a specific result of a specific series.

        Args:
            series_id (str): The ID of the series.
            result_id (str): The ID of the result.

        Returns:
            dict: The result set returned.
        """

        path = "/".join([self._PREFIX, series_id, result_id])
        return self._get(path)



"""
Base for interacting with the Censys ASM API.
"""
# pylint: disable=too-many-arguments
import os

from math import inf
from typing import Generator, Type, Optional
from requests.models import Response
from censys.exceptions import (
    CensysMissingApiKeyException,
    CensysAsmException,
    CensysExceptionMapper,
)

from censys.base import CensysAPIBase
from censys.config import get_config, DEFAULT


class CensysAsmAPI(CensysAPIBase):
    """
    This is the base class for ASM's Seeds, Assets, and Events classes
    """

    DEFAULT_URL: str = "https://app.censys.io/api/v1"
    """Default ASM API base URL."""

    def __init__(self, api_key: Optional[str] = None, **kwargs):
        url = kwargs.pop("url", self.DEFAULT_URL)
        CensysAPIBase.__init__(self, url=url, **kwargs)

        # Gets config file
        config = get_config()

        # Try to get credentials
        self._api_key = (
            api_key
            or os.getenv("CENSYS_ASM_API_KEY")
            or config.get(DEFAULT, "asm_api_key")
        )

        if not self._api_key:
            raise CensysMissingApiKeyException("No ASM API key configured.")

        self._session.headers.update(
            {"Content-Type": "application/json", "Censys-Api-Key": self._api_key}
        )

    def _get_exception_class(  # type: ignore
        self, res: Response
    ) -> Type[CensysAsmException]:
        return CensysExceptionMapper.ASM_EXCEPTIONS.get(
            res.json()["errorCode"], CensysAsmException
        )

    def _get_page(
        self, path: str, page_number: int = 1, page_size: Optional[int] = None
    ) -> Generator[dict, None, None]:
        """
        Fetches paginated ASM resource API results.

        Args:
            path (str): The API url endpoint.
            page_number (int, optional): Page number to begin at when getting results.
            page_size (int, optional): Number of results to return per HTTP request

        Returns:
            generator: The resource result set returned.
        """

        total_pages = inf

        while page_number <= total_pages:
            args = {"pageNumber": page_number, "pageSize": page_size or 500}

            res = self._get(path, args=args)
            page_number = int(res["pageNumber"]) + 1
            total_pages = int(res["totalPages"])

            if "comments" in path:
                for comment in res["comments"]:
                    yield comment
            else:
                for asset in res["assets"]:
                    yield asset

    def _get_logbook_page(
        self, path: str, args: Optional[dict] = None
    ) -> Generator[dict, None, None]:
        """
        Fetches paginated ASM logbook API events.

        Args:
            path (str): The API url endpoint.
            args (dict, optional): URL args that are mapped to params (cursor).

        Returns:
            generator: The event result set returned.
        """

        end_of_events = False

        while not end_of_events:
            res = self._get(path, args=args)
            end_of_events = res["endOfEvents"]
            args = {"cursor": res["nextCursor"]}

            for event in res["events"]:
                yield event



"""
Base for interacting with the Censys API's.
"""

import os
import json
import warnings
from functools import wraps
from typing import Any, Callable, List, Optional, Type

import backoff
import requests
from requests.models import Response

from censys import __name__ as NAME
from censys import __version__ as VERSION
from censys.exceptions import (
    CensysAPIException,
    CensysException,
    CensysJSONDecodeException,
    CensysRateLimitExceededException,
    CensysTooManyRequestsException,
)

Fields = Optional[List[str]]


# Wrapper to make max_retries configurable at runtime
def _backoff_wrapper(method: Callable):
    @wraps(method)
    def _wrapper(self, *args, **kwargs):
        @backoff.on_exception(
            backoff.expo,
            (
                CensysRateLimitExceededException,
                CensysTooManyRequestsException,
            ),
            max_tries=self.max_retries,
        )
        def _impl():
            return method(self, *args, *kwargs)

        return _impl()

    return _wrapper


class CensysAPIBase:
    """
    This is the base class for API queries.

    Args:
        url (str, optional): The URL to make API requests.
        timeout (int, optional): Timeout for API requests in seconds.
        user_agent (str, optional): Override User-Agent string.
        proxies (dict, optional): Configure HTTP proxies.

    Raises:
        CensysException: Base Exception Class for the Censys API.
    """

    DEFAULT_TIMEOUT: int = 30
    """Default API timeout."""
    DEFAULT_USER_AGENT: str = "%s/%s" % (NAME, VERSION)
    """Default API user agent."""
    DEFAULT_MAX_RETRIES: int = 10
    """Default max number of API retries."""

    def __init__(self, url: Optional[str] = None, **kwargs):
        # Get common request settings
        self.timeout = kwargs.get("timeout") or self.DEFAULT_TIMEOUT
        self.max_retries = kwargs.get("max_retries") or self.DEFAULT_MAX_RETRIES
        self._api_url = url or os.getenv("CENSYS_API_URL")

        if not self._api_url:
            raise CensysException("No API url configured.")

        # Create a session and set credentials
        self._session = requests.Session()
        proxies = kwargs.get("proxies")
        if proxies:
            if "http" in proxies.keys():
                warnings.warn("HTTP proxies will not be used.")
                proxies.pop("http", None)
            self._session.proxies = proxies
        self._session.headers.update(
            {
                "accept": "application/json, */8",
                "User-Agent": " ".join(
                    [
                        requests.utils.default_user_agent(),
                        kwargs.get("user_agent")
                        or kwargs.get("user_agent_identifier")
                        or self.DEFAULT_USER_AGENT,
                    ]
                ),
            }
        )

    @staticmethod
    def _get_exception_class(_: Response) -> Type[CensysAPIException]:
        """
        Maps HTTP status code or ASM error code to exception.
        Must be implemented by child class.

        Args:
            _ (Response): HTTP requests response object.

        Returns:
            Type[CensysAPIException]: Exception to raise.
        """
        return CensysAPIException

    @_backoff_wrapper
    def _make_call(
        self,
        method: Callable,
        endpoint: str,
        args: Optional[dict] = None,
        data: Optional[Any] = None,
    ) -> dict:
        """
        Wrapper functions for all our REST API calls checking for errors
        and decoding the responses.

        Args:
            method (Callable): Method to send HTTP request.
            endpoint (str): The path of API endpoint.
            args (dict, optional): URL args that are mapped to params.
            data (Any, optional): JSON data to serialize with request.

        Raises:
            CensysJSONDecodeException: The response is not valid JSON.

        Returns:
            dict: Results from an API request.
        """

        if endpoint.startswith("/"):
            url = f"{self._api_url}{endpoint}"
        else:
            url = f"{self._api_url}/{endpoint}"

        request_kwargs = {
            "params": args or {},
            "timeout": self.timeout,
        }

        if data:
            data = json.dumps(data)
            request_kwargs["data"] = data

        res = method(url, **request_kwargs)

        if res.status_code == 200:
            # Check for a returned json body
            try:
                return res.json()
            # Successful request returned no json body in response
            except ValueError:
                return {}

        try:
            json_data = res.json()
            message = json_data.get("error") or json_data["message"]
            const = json_data.get("error_type", None)
            error_code = json_data.get("errorCode", None)
            details = json_data.get("details", None)
        except (ValueError, json.decoder.JSONDecodeError) as error:  # pragma: no cover
            message = (
                f"Response from {res.url} is not valid JSON and cannot be decoded."
            )
            raise CensysJSONDecodeException(
                status_code=res.status_code,
                message=message,
                body=res.text,
                const="badjson",
            ) from error
        except KeyError:  # pragma: no cover
            message = None
            const = "unknown"
            details = "unknown"
            error_code = "unknown"

        censys_exception = self._get_exception_class(res)
        raise censys_exception(
            status_code=res.status_code,
            body=res.text,
            const=const,
            message=message,
            error_code=error_code,
            details=details,
        )

    def _get(self, endpoint: str, args: Optional[dict] = None) -> dict:
        return self._make_call(self._session.get, endpoint, args)

    def _post(
        self, endpoint: str, args: Optional[dict] = None, data: Optional[dict] = None
    ) -> dict:
        return self._make_call(self._session.post, endpoint, args, data)

    def _put(
        self, endpoint: str, args: Optional[dict] = None, data: Optional[dict] = None
    ) -> dict:
        return self._make_call(
            self._session.put, endpoint, args, data
        )  # pragma: no cover

    def _delete(self, endpoint: str, args: Optional[dict] = None) -> dict:
        return self._make_call(self._session.delete, endpoint, args)  # pragma: no cover



"""
Interact with the Censys Search Website API.
"""

from censys.api import CensysSearchAPI


class CensysWebsites(CensysSearchAPI):
    """
    Interacts with the Websites index.
    """

    INDEX_NAME = "websites"
    """Name of Censys Index."""



#!/usr/bin/env python
"""
Interact with the Censys Search API through the command line.
"""

import os
import sys
import csv
import time
import json
import argparse
from pathlib import Path
from typing import Union, Optional, List, Tuple

import requests

from censys.api import CensysSearchAPI
from censys.asm.client import AsmClient
from censys.config import get_config, write_config, DEFAULT
from censys.ipv4 import CensysIPv4
from censys.websites import CensysWebsites
from censys.certificates import CensysCertificates
from censys.exceptions import (
    CensysCLIException,
    CensysNotFoundException,
    CensysUnauthorizedException,
)

Fields = List[str]
Results = List[dict]
Index = Union[CensysIPv4, CensysWebsites, CensysCertificates]


# TODO: Breakout into cli/
class CensysAPISearch:
    """
    This class searches the Censys API, taking in options from the command line and
    returning the results to a CSV or JSON file, or to stdout.

    Args:
        api_id (str, optional): The API ID provided by Censys.
        api_secret (str, optional): The API secret provided by Censys.
        start_page (int, optional): Page number to start from. Defaults to 1.
        max_pages (int, optional): The maximum number of pages. Defaults to 10.
    """

    csv_fields: Fields = list()
    """A list of fields to be used by the CSV writer."""

    def __init__(self, **kwargs):
        self.api_user = kwargs.get("api_id")
        self.api_pass = kwargs.get("api_secret")
        self.start_page = kwargs.get("start_page", 1)
        self.max_pages = kwargs.get("max_pages", 10)

    @staticmethod
    def _write_csv(file_path: str, search_results: Results, fields: Fields) -> bool:
        """
        This method writes the search results to a new file in CSV format.

        Args:
            file_path (str): Name of the file to write to on the disk.
            search_results (Results): A list of results from the query.
            fields (Fields): A list of fields to write as headers.

        Returns:
            bool: True if wrote to file successfully.
        """

        with open(file_path, "w") as output_file:
            if search_results and isinstance(search_results, list):
                # Get the header row from the first result
                writer = csv.DictWriter(output_file, fieldnames=fields)
                writer.writeheader()

                for result in search_results:
                    # Use the Dict writer to process and write results to CSV
                    writer.writerow(result)

        print(f"Wrote results to file {file_path}")

        # method returns True, if the file has been written successfully.
        return True

    @staticmethod
    def _write_json(file_path: str, search_results: Results) -> bool:
        """
        This method writes the search results to a new file in JSON format.

        Args:
            file_path (str): Name of the file to write to on the disk.
            search_results (Results): A list of results from the query.

        Returns:
            bool: True if wrote to file successfully.
        """

        with open(file_path, "w") as output_file:
            # Since the results are already in JSON, just write them to a file.
            json.dump(search_results, output_file, indent=4)

        print(f"Wrote results to file {file_path}")
        return True

    @staticmethod
    def _write_screen(search_results: Results) -> bool:
        """
        This method writes the search results to screen.

        Args:
            search_results (Results): A list of results from the query.

        Returns:
            bool: True if wrote to file successfully.
        """

        print(json.dumps(search_results, indent=4))
        return True

    def write_file(
        self,
        results_list: Results,
        file_format: str = "screen",
        file_path: Optional[str] = None,
    ) -> bool:
        """
        This method just sorts which format will be used to store
        the results of the query.

        Args:
            results_list (Results): A list of results from the API query.
            file_format (str, optional): The format of the output.
            file_path (str optional): A path to write results to.

        Returns:
            bool: True if wrote out successfully.
        """

        if file_format and isinstance(file_format, str):
            file_format = file_format.lower()

        if not file_path:
            # This method just creates some dynamic file names
            file_name_ext = f"{time.time()}.{file_format}"
            file_path = f"censys-query-output.{file_name_ext}"

        if file_format == "json":
            return self._write_json(file_path, results_list)
        if file_format == "csv":
            return self._write_csv(file_path, results_list, fields=self.csv_fields)
        return self._write_screen(results_list)

    def _combine_fields(
        self,
        default_fields: Fields,
        user_fields: Fields,
        overwrite: bool = False,
    ) -> Fields:
        """
        This method is used to specify which fields will be returned in the results.

        Args:
            default_fields (Fields): A list of fields that are returned by default.
            user_fields (Fields): A list of user-specified fields. Max 20.
            overwrite (bool, optional): Whether to overwrite or append default fields
                                        with user fields. Defaults to False.

        Raises:
            CensysCLIException: Too many fields specified.

        Returns:
            Fields: A list of fields.
        """

        field_list: Fields = default_fields

        if user_fields:
            if overwrite:
                field_list = user_fields
            else:
                field_list = list(set(user_fields + default_fields))

        # This is the hard limit for the number of fields that can be in a query.
        if len(list(field_list)) > 20:
            raise CensysCLIException(
                "Too many fields specified. The maximum number of fields is 20."
            )

        self.csv_fields = list(field_list)
        return list(field_list)

    def _process_search(
        self, query: str, search_index: Index, fields: Fields
    ) -> Results:
        """
        This method provides a common way to process searches from the API.

        Args:
            query (str): The string to send to the API as a query.
            search_index (Index): The data set to be queried.
            fields (Fields): A list of fields to be returned for each result.

        Returns:
            Results: A list of results from the query.
        """

        records = []

        while True:
            response = search_index.paged_search(
                query=query, fields=fields, page=self.start_page
            )

            for record in response["results"]:
                records.append(record)

            # Break while loop when last page is reached
            if (
                response["metadata"]["page"] >= response["metadata"]["pages"]
                or response["metadata"]["page"] >= self.max_pages
            ):
                break

            self.start_page += 1

        return records

    def search_ipv4(self, **kwargs) -> Results:
        """
        A method to search the IPv4 data set via the API.

        Args:
            query (str): The string search query.
            fields (list, optional): The fields that should be returned with a query.
            overwrite (bool, optional): Whether to overwrite or append default fields
                                        with user fields. Defaults to False.

        Returns:
            Results: A list of results from the query.
        """

        default_fields = [
            "updated_at",
            "protocols",
            "metadata.description",
            "autonomous_system.name",
            "23.telnet.banner.banner",
            "80.http.get.title",
            "80.http.get.metadata.description",
            "8080.http.get.metadata.description",
            "8888.http.get.metadata.description",
            "443.https.get.metadata.description",
            "443.https.get.title",
            "443.https.tls.certificate.parsed.subject_dn",
            "443.https.tls.certificate.parsed.names",
            "443.https.tls.certificate.parsed.subject.common_name",
            "443.https.tls.certificate.parsed.extensions.subject_alt_name.dns_names",
        ]

        query = kwargs.get("query", "")
        fields = kwargs.get("fields", [])
        overwrite = kwargs.get("overwrite", False)

        index = CensysIPv4(api_id=self.api_user, api_secret=self.api_pass)

        return self._process_search(
            query,
            index,
            self._combine_fields(default_fields, fields, overwrite=overwrite),
        )

    def search_certificates(self, **kwargs) -> Results:
        """
        A method to search the Certificates data set via the API.

        Args:
            query (str): The string search query.
            fields (list, optional): The fields that should be returned with a query.
            overwrite (bool, optional): Whether to overwrite or append default fields
                                        with user fields. Defaults to False.

        Returns:
            Results: A list of results from the query.
        """

        default_fields = [
            "metadata.updated_at",
            "parsed.issuer.common_name",
            "parsed.names",
            "parsed.serial_number",
            "parsed.self_signed",
            "parsed.subject.common_name",
            "parsed.validity.start",
            "parsed.validity.end",
            "parsed.validity.length",
            "metadata.source",
            "metadata.seen_in_scan",
            "tags",
        ]

        query = kwargs.get("query", "")
        fields = kwargs.get("fields", [])
        overwrite = kwargs.get("overwrite", False)

        index = CensysCertificates(api_id=self.api_user, api_secret=self.api_pass)

        return self._process_search(
            query,
            index,
            self._combine_fields(default_fields, fields, overwrite=overwrite),
        )

    def search_websites(self, **kwargs) -> Results:
        """
        A method to search the Websites (Alexa Top 1M) data set via the API.

        Args:
            query (str): The string search query.
            fields (list, optional): The fields that should be returned with a query.
            overwrite (bool, optional): Whether to overwrite or append default fields
                                        with user fields. Defaults to False.

        Returns:
            Results: A list of results from the query.
        """

        default_fields = [
            "443.https.tls.version",
            "alexa_rank",
            "domain",
            "ports",
            "protocols",
            "tags",
            "updated_at",
        ]

        query = kwargs.get("query", "")
        fields = kwargs.get("fields", [])
        overwrite = kwargs.get("overwrite", False)

        index = CensysWebsites(api_id=self.api_user, api_secret=self.api_pass)

        return self._process_search(
            query,
            index,
            self._combine_fields(default_fields, fields, overwrite=overwrite),
        )


class CensysHNRI:
    """
    This class searches the Censys API, check the user's current IP for risks.

    Args:
        api_id (str, optional): The API ID provided by Censys.
        api_secret (str, optional): The API secret provided by Censys.
    """

    HIGH_RISK_DEFINITION: List[str] = ["telnet", "redis", "postgres", "vnc"]
    MEDIUM_RISK_DEFINITION: List[str] = ["ssh", "http", "https"]

    def __init__(self, api_id: str, api_secret: str):
        self.index = CensysIPv4(api_id, api_secret)

    @staticmethod
    def get_current_ip() -> str:
        """
        Uses ipify.org to get the current IP address.

        Returns:
            str: IP address.
        """

        response = requests.get("https://api.ipify.org?format=json")
        current_ip = response.json().get("ip")
        return current_ip

    def translate_risk(self, protocols: list) -> Tuple[list, list]:
        """
        Interpret protocols to risks.

        Args:
            protocols (list): List of slash divided ports/protocols.

        Returns:
            Tuple[list, list]: Lists of high and medium risks.
        """

        high_risk = []
        medium_risk = []

        for protocol in protocols:
            port, protocol = protocol.split("/")
            string = f"{protocol} on {port}"
            if protocol in self.HIGH_RISK_DEFINITION:
                high_risk.append({"port": port, "protocol": protocol, "string": string})
            elif protocol in self.MEDIUM_RISK_DEFINITION:
                medium_risk.append(
                    {"port": port, "protocol": protocol, "string": string}
                )
            elif protocol == "banner":
                medium_risk.append(
                    {"port": port, "protocol": "unknown protocol", "string": string}
                )
            else:
                medium_risk.append(
                    {"port": port, "protocol": protocol, "string": string}
                )

        return high_risk, medium_risk

    @staticmethod
    def risks_to_string(high_risk: list, medium_risk: list) -> str:
        """
        Risks to printable string.

        Args:
            high_risk (list): Lists of high risks.
            medium_risk (list): Lists of medium risks.

        Raises:
            CensysCLIException: No information/risks found.

        Returns:
            str: Printable string for CLI.
        """

        len_high_risk = len(high_risk)
        len_medium_risk = len(medium_risk)

        if len_high_risk + len_medium_risk == 0:
            raise CensysCLIException

        response = ""
        if len_high_risk > 0:
            response = (
                response
                + "High Risks Found: \n"
                + "\n".join([risk.get("string") for risk in high_risk])
            )
        else:
            response = response + "You don't have any High Risks in your network\n"
        if len_medium_risk > 0:
            response = (
                response
                + "Medium Risks Found: \n"
                + "\n".join([risk.get("string") for risk in medium_risk])
            )
        else:
            response = response + "You don't have any Medium Risks in your network\n"
        return response

    def view_current_ip_risks(self) -> str:
        """
        Gets protocol information for the current IP and returns any risks.

        Returns:
            str: Printable
        """

        current_ip = self.get_current_ip()

        try:
            results = self.index.view(current_ip)
            protocols = results.get("protocols", [])
            high_risk, medium_risk = self.translate_risk(protocols)
            return self.risks_to_string(high_risk, medium_risk)
        except (CensysNotFoundException, CensysCLIException):
            return "No Risks were found on your network"


def search(args):
    """
    search subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """

    censys_args = {}

    if args.start_page:
        censys_args["start_page"] = args.start_page

    if args.max_pages:
        censys_args["max_pages"] = args.max_pages

    if args.api_id:
        censys_args["api_id"] = args.api_id

    if args.api_secret:
        censys_args["api_secret"] = args.api_secret

    censys = CensysAPISearch(**censys_args)

    search_args = {"query": args.query}

    if args.fields:
        search_args["fields"] = args.fields

    if args.overwrite:
        search_args["overwrite"] = args.overwrite

    indexes = {
        "ipv4": censys.search_ipv4,
        "certs": censys.search_certificates,
        "websites": censys.search_websites,
    }

    index_type = args.index_type or args.query_type

    index_func = indexes[index_type]
    results = index_func(**search_args)

    try:
        censys.write_file(results, file_format=args.format, file_path=args.output)
    except ValueError as error:  # pragma: no cover
        print(f"Error writing log file. Error: {error}")


def hnri(args):
    """
    hnri subcommand.

    Args:
        args (Namespace): Argparse Namespace.
    """

    client = CensysHNRI(args.api_id, args.api_secret)

    risks = client.view_current_ip_risks()

    print(risks)


def cli_config(_):  # pragma: no cover
    """
    config subcommand.

    Args:
        _: Argparse Namespace.
    """

    api_id_prompt = "Censys API ID"
    api_secret_prompt = "Censys API Secret"

    config = get_config()
    api_id = config.get(DEFAULT, "api_id")
    api_secret = config.get(DEFAULT, "api_secret")

    if api_id and api_secret:
        redacted_id = api_id.replace(api_id[:32], 32 * "*")
        redacted_secret = api_secret.replace(api_secret[:28], 28 * "*")
        api_id_prompt = f"{api_id_prompt} [{redacted_id}]"
        api_secret_prompt = f"{api_secret_prompt} [{redacted_secret}]"

    api_id = input(api_id_prompt + ": ").strip() or api_id
    api_secret = input(api_secret_prompt + ": ").strip() or api_secret

    if not (api_id and api_secret):
        print("Please enter valid credentials")
        sys.exit(1)

    try:
        client = CensysSearchAPI(api_id, api_secret)
        account = client.account()
        email = account.get("email")

        # Assumes that login was successfully
        config.set(DEFAULT, "api_id", api_id)
        config.set(DEFAULT, "api_secret", api_secret)

        write_config(config)
        print(f"\nSuccessfully authenticated for {email}")
        sys.exit(0)
    except CensysUnauthorizedException:
        print("Failed to authenticate")
        sys.exit(1)


def cli_asm_config(_):  # pragma: no cover
    """
    config asm subcommand.

    Args:
        _: Argparse Namespace.
    """

    api_key_prompt = "Censys ASM API Key"

    config = get_config()
    api_key = config.get(DEFAULT, "asm_api_key")

    if api_key:
        key_len = len(api_key) - 4
        redacted_api_key = api_key.replace(api_key[:key_len], key_len * "*")
        api_key_prompt = f"{api_key_prompt} [{redacted_api_key}]"

    api_key = input(api_key_prompt + ": ").strip() or api_key

    if not api_key:
        print("Please enter valid credentials")
        sys.exit(1)

    try:
        AsmClient(api_key)

        # Assumes that login was successfully
        config.set(DEFAULT, "asm_api_key", api_key)

        write_config(config)
        print("\nSuccessfully authenticated")
        sys.exit(0)
    except CensysUnauthorizedException:
        print("Failed to authenticate")
        sys.exit(1)


def get_parser() -> argparse.ArgumentParser:
    """
    Gets ArgumentParser for CLI.

    Returns:
        argparse.ArgumentParser
    """

    config = get_config()

    auth = argparse.ArgumentParser(add_help=False)
    auth.add_argument(
        "--api-id",
        default=os.getenv("CENSYS_API_ID") or config.get(DEFAULT, "api_id"),
        required=False,
        help="a Censys API ID \
            (alternatively you can use the env variable CENSYS_API_ID)",
    )
    auth.add_argument(
        "--api-secret",
        default=os.getenv("CENSYS_API_SECRET") or config.get(DEFAULT, "api_secret"),
        required=False,
        help="a Censys API SECRET \
            (alternatively you can use the env variable CENSYS_API_SECRET)",
    )

    parser = argparse.ArgumentParser()
    parser.set_defaults()
    subparsers = parser.add_subparsers()

    # Search Specific Args
    search_parser = subparsers.add_parser(
        "search",
        description="Query Censys Search for resource data by providing a query \
            string, the resource index, and the fields to be returned",
        help="query Censys search",
        parents=[auth],
    )
    search_parser.add_argument(
        "-q",
        "--query",
        type=str,
        required=True,
        help="a string written in Censys Search syntax",
    )

    index_types = ["ipv4", "certs", "websites"]
    index_metavar = "ipv4|certs|websites"
    index_default = "ipv4"
    search_parser.add_argument(
        "--index-type",
        type=str,
        default=index_default,
        choices=index_types,
        metavar=index_metavar,
        help="which resource index to query",
    )
    # Backwards compatibility
    search_parser.add_argument(
        "--query_type",
        type=str,
        default=index_default,
        choices=index_types,
        metavar=index_metavar,
        help=argparse.SUPPRESS,
    )

    search_parser.add_argument(
        "--fields", nargs="+", help="list of index-specific fields"
    )
    search_parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="overwrite instead of append fields returned by default \
            with fields provided in the fields argument",
    )
    search_parser.add_argument(
        "-f",
        "--format",
        type=str,
        default="screen",
        metavar="json|csv|screen",
        help="format of output",
    )
    search_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="output file path",
    )
    search_parser.add_argument(
        "--start-page", default=1, type=int, help="page number to start from"
    )
    search_parser.add_argument(
        "--max-pages",
        default=1,
        type=int,
        help="maximum number of pages of results to return",
    )
    search_parser.set_defaults(func=search)

    # HNRI Specific Args
    hnri_parser = subparsers.add_parser(
        "hnri",
        description="Home Network Risk Identifier (H.N.R.I.)",
        help="home network risk identifier",
        parents=[auth],
    )
    hnri_parser.set_defaults(func=hnri)

    # Config Specific Args
    config_parser = subparsers.add_parser(
        "config",
        description="Configure Censys Search API Settings",
        help="configure Censys search API settings",
    )
    config_parser.set_defaults(func=cli_config)

    # ASM Config Args
    asm_config_parser = subparsers.add_parser(
        "config-asm",
        description="Configure Censys ASM API Settings",
        help="configure Censys ASM API settings",
    )
    asm_config_parser.set_defaults(func=cli_asm_config)

    return parser


def main():
    """main cli function"""

    parser = get_parser()

    # Executes by subcommand
    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError:
        parser.print_help()
        parser.exit()
    except KeyboardInterrupt:  # pragma: no cover
        sys.exit(1)


if __name__ == "__main__":  # pragma: no cover
    main()



"""
Base for interacting with the Censys Search API.
"""

import os
from typing import Generator, List, Optional, Type

from requests.models import Response

from censys.base import CensysAPIBase
from censys.config import DEFAULT, get_config
from censys.exceptions import (
    CensysException,
    CensysSearchException,
    CensysExceptionMapper,
)

Fields = Optional[List[str]]


class CensysSearchAPI(CensysAPIBase):
    """
    This class is the base class for the Data, Certificate, IPv4, and Website index.
    """

    DEFAULT_URL: str = "https://censys.io/api/v1"
    """Default Search API base URL."""
    INDEX_NAME: Optional[str] = None
    """Name of Censys Index."""

    def __init__(
        self,
        api_id: Optional[str] = None,
        api_secret: Optional[str] = None,
        url: Optional[str] = DEFAULT_URL,
        **kwargs,
    ):
        CensysAPIBase.__init__(self, url, **kwargs)

        # Gets config file
        config = get_config()

        # Try to get credentials
        self._api_id = (
            api_id or os.getenv("CENSYS_API_ID") or config.get(DEFAULT, "api_id")
        )
        self._api_secret = (
            api_secret
            or os.getenv("CENSYS_API_SECRET")
            or config.get(DEFAULT, "api_secret")
        )
        if not self._api_id or not self._api_secret:
            raise CensysException("No API ID or API secret configured.")

        self._session.auth = (self._api_id, self._api_secret)

        # Generate concrete paths to be called
        self.search_path = f"search/{self.INDEX_NAME}"
        self.view_path = f"view/{self.INDEX_NAME}"
        self.report_path = f"report/{self.INDEX_NAME}"

        # Confirm setup
        self.account()

    def _get_exception_class(  # type: ignore
        self, res: Response
    ) -> Type[CensysSearchException]:
        return CensysExceptionMapper.SEARCH_EXCEPTIONS.get(
            res.status_code, CensysSearchException
        )

    def account(self) -> dict:
        """
        Gets the current account information. Including email and quota.

        Returns:
            dict: Account response.
        """

        return self._get("account")

    def quota(self) -> dict:
        """
        Gets the current account's query quota.

        Returns:
            dict: Quota response.
        """

        return self.account()["quota"]

    def metadata(self, query: str) -> dict:
        """
        Returns metadata of a given search query.

        Args:
            query (str): The query to be executed.

        Returns:
            dict: The metadata of the result set returned.
        """

        data = {"query": query, "page": 1, "fields": []}
        return self._post(self.search_path, data=data).get("metadata", {})

    def paged_search(
        self,
        query: str,
        fields: Fields = None,
        page: int = 1,
        flatten: bool = True,
    ) -> dict:
        """
        Searches the given index for all records that match the given query.

        Args:
            query (str): The query to be executed.
            fields (Fields, optional): Fields to be returned in the result set.
            page (int, optional): The page of the result set. Defaults to 1.
            flatten (bool, optional): Flattens fields to dot notation. Defaults to True.

        Returns:
            dict: The result set returned.
        """

        page = int(page)
        data = {
            "query": query,
            "page": page,
            "fields": fields or [],
            "flatten": flatten,
        }
        return self._post(self.search_path, data=data)

    # pylint: disable=too-many-arguments
    def search(
        self,
        query: str,
        fields: Fields = None,
        page: int = 1,
        max_records: Optional[int] = None,
        flatten: bool = True,
    ) -> Generator[dict, None, None]:
        """
        Searches the given index for all records that match the given query.
        For more details, see our documentation: https://censys.io/api/v1/docs/search

        Args:
            query (str): The query to be executed.
            fields (Fields, optional): Fields to be returned in the result set.
            page (int, optional): The page of the result set. Defaults to 1.
            max_records (Optional[int], optional): The maximum number of records.
            flatten (bool, optional): Flattens fields to dot notation. Defaults to True.

        Yields:
            dict: The result set returned.
        """

        if fields is None:
            fields = []
        page = int(page)
        pages = float("inf")
        data = {"query": query, "page": page, "fields": fields, "flatten": flatten}

        count = 0
        while page <= pages:
            payload = self._post(self.search_path, data=data)
            pages = payload["metadata"]["pages"]
            page += 1
            data["page"] = page

            for result in payload["results"]:
                yield result
                count += 1
                if max_records and count >= max_records:
                    return

    def view(self, document_id: str) -> dict:
        """
        View the current structured data we have on a specific document.
        For more details, see our documentation: https://censys.io/api/v1/docs/view

        Args:
            document_id (str): The ID of the document you are requesting.

        Returns:
            dict: The result set returned.
        """

        return self._get("/".join((self.view_path, document_id)))

    def report(self, query: str, field: str, buckets: int = 50) -> dict:
        """
        Creates a report on the breakdown of the values of a field in a result set.
        For more details, see our documentation: https://censys.io/api/v1/docs/report

        Args:
            query (str): The query to be executed.
            field (str): The field you are running a breakdown on.
            buckets (int, optional): The maximum number of values. Defaults to 50.

        Returns:
            dict: The result set returned.
        """

        data = {"query": query, "field": field, "buckets": int(buckets)}
        return self._post(self.report_path, data=data)



"""
Interact with the Censys Seeds, Assets, and Logbook APIs.
"""

from typing import Optional

from censys.asm.seeds import Seeds
from censys.asm.assets import Assets
from censys.asm.events import Events


class AsmClient:
    """
    Client ASM API class.
    """

    def __init__(self, api_key: Optional[str] = None, **kwargs):
        self.seeds = Seeds(api_key, **kwargs)
        self.hosts = Assets("hosts", api_key, **kwargs)
        self.certificates = Assets("certificates", api_key, **kwargs)
        self.domains = Assets("domains", api_key, **kwargs)
        self.events = Events(api_key, **kwargs)



import io
import os
from setuptools import setup, find_packages

import censys

NAME = "censys"
DESCRIPTION = censys.__doc__.strip()
GIT_URL = "https://github.com/censys/censys-python"
ISSUE_URL = GIT_URL + "/issues"
DOC_URL = "https://censys-python.rtfd.io"


try:
    here = os.path.abspath(os.path.dirname(__file__))
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

setup(
    name=NAME,
    version=censys.__version__,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=censys.__author__,
    author_email=censys.__email__,
    license=censys.__license__,
    url=GIT_URL,
    keywords=NAME,
    python_requires=">=3.6.0",
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=["requests", "backoff"],
    extras_require={
        "dev": [
            "flake8",
            "black",
            "pytest",
            "pytest-cov",
            "requests-mock",
            "mypy",
            "backoff-stubs",
            "twine",
            "parameterized",
        ],
        "docs": ["sphinx"],
    },
    entry_points={"console_scripts": ["censys = censys.cli:main"]},
    classifiers=[
        "Typing :: Typed",
        "Topic :: Internet",
        "Topic :: Security",
        "Topic :: Documentation :: Sphinx",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: Pytest",
        "Framework :: Flake8",
        "Environment :: Console",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    project_urls={"Documentation": DOC_URL, "Source": GIT_URL, "Tracker": ISSUE_URL},
)



"""
Interact with the Censys Assets API.
"""

from typing import Generator, Optional

from censys.asm.api import CensysAsmAPI


class Assets(CensysAsmAPI):
    """
    Assets API class
    """

    def __init__(self, asset_type, api_key: Optional[str] = None, **kwargs):
        CensysAsmAPI.__init__(self, api_key, **kwargs)
        self.base_path = f"assets/{asset_type}"

    def get_assets(
        self, page_number: int = 1, page_size: Optional[int] = None
    ) -> Generator[dict, None, None]:
        """
        Requests assets data.

        Args:
            page_number (int, optional): Page number to begin at when searching.
            page_size (int, optional): Page size for retrieving assets.

        Returns:
            generator: Asset search results.
        """

        return self._get_page(
            self.base_path, page_number=page_number, page_size=page_size
        )

    def get_asset_by_id(self, asset_id: str) -> dict:
        """
        Requests asset data by ID.

        Args:
            asset_id (str): Requested asset ID.

        Returns:
            dict: Asset search result.
        """

        path = f"{self.base_path}/{asset_id}"

        return self._get(path)

    def get_comments(
        self,
        asset_id: str,
        page_number: int = 1,
        page_size: Optional[int] = None,
    ) -> Generator[dict, None, None]:
        """
        Requests comments on a specified asset.

        Args:
            asset_id (str): Asset ID for requested comments.
            page_number (int, optional): Page number to begin at when searching.
            page_size (int, optional): Page size for retrieving comments.

        Returns:
            generator: Comment search results.
        """

        path = f"{self.base_path}/{asset_id}/comments"

        return self._get_page(path, page_number=page_number, page_size=page_size)

    def get_comment_by_id(self, asset_id: str, comment_id: int) -> dict:
        """
        Requests a comment on a specified asset by comment ID.

        Args:
            asset_id (str): Asset ID for requested comments.
            comment_id (int): Requested comment ID.

        Returns:
            dict: Comment search result.
        """

        path = f"{self.base_path}/{asset_id}/comments/{comment_id}"

        return self._get(path)

    def add_comment(self, asset_id: str, comment: str) -> dict:
        """
        Adds a comment to a specified asset on the ASM platform.

        Args:
            asset_id (str): Asset ID to add comment to.
            comment (str): New comment text.

        Returns:
            dict: Added comment results.
        """

        path = f"{self.base_path}/{asset_id}/comments"
        data = {"markdown": str(comment)}

        return self._post(path, data=data)

    def add_tag(self, asset_id: str, name: str, color: Optional[str] = None) -> dict:
        """
        Adds a tag to a specified asset on the ASM platform.

        Args:
            asset_id (str): Asset ID to add tag to.
            name (str): New tag name.
            color (str, optional): New tag color.
        """

        path = f"{self.base_path}/{asset_id}/tags"
        data = format_tag(name, color)

        return self._post(path, data=data)

    def delete_tag(self, asset_id: str, name: str) -> dict:
        """
        Deletes a tag from a specified asset on the ASM platform by tag name.

        Args:
            asset_id (str): Asset ID to delete tag from.
            name (str): Tag name to delete.
        """

        path = f"{self.base_path}/{asset_id}/tags/{name}"

        return self._delete(path)


def format_tag(name: str, color: Optional[str] = None) -> dict:
    """
    Formats tag name and color request data.

    Args:
        name (str): Tag name.
        color (str, optional): Tag color.

    Returns:
            dict: Formatted tag request data.
    """

    if color:
        return {"name": str(name), "color": str(color)}

    return {"name": str(name)}
